@@ -17,9 +17,10 @@
 
 #include "mozilla/gfx/gfxVars.h"
 #include "mozilla/Preferences.h"
+#include "mozilla/ScopeExit.h"
+#include "mozilla/StaticPrefs_gl.h"
 #include "mozilla/StaticPtr.h"
 #include "mozilla/layers/CompositorOptions.h"
-#include "mozilla/webrender/RenderThread.h"
 #include "mozilla/widget/CompositorWidget.h"
 #include "mozilla/widget/WinCompositorWidget.h"
 
@@ -93,13 +94,11 @@ bool WGLLibrary::EnsureInitialized() {
     }
   }
 
-#define SYMBOL(X)                 \
-  {                               \
-    (PRFuncPtr*)&mSymbols.f##X, { \
-      { "wgl" #X }                \
-    }                             \
+#define SYMBOL(X) {(PRFuncPtr*)&mSymbols.f##X, {{"wgl" #X}}}
+#define END_OF_SYMBOLS \
+  {                    \
+    nullptr, {}        \
   }
-#define END_OF_SYMBOLS {nullptr, {}}
 
   {
     const auto loader = SymbolLoader(*mOGLLibrary);
@@ -193,7 +192,6 @@ bool WGLLibrary::EnsureInitialized() {
 
   const auto extString = mSymbols.fGetExtensionsStringARB(mRootDc);
   MOZ_ASSERT(extString);
-  MOZ_ASSERT(HasExtension(extString, "WGL_ARB_extensions_string"));
 
   // --
 
@@ -265,20 +263,18 @@ void WGLLibrary::Reset() {
   }
 }
 
-GLContextWGL::GLContextWGL(CreateContextFlags flags, const SurfaceCaps& caps,
-                           bool isOffscreen, HDC aDC, HGLRC aContext,
+GLContextWGL::GLContextWGL(const GLContextDesc& desc, HDC aDC, HGLRC aContext,
                            HWND aWindow)
-    : GLContext(flags, caps, nullptr, isOffscreen),
+    : GLContext(desc, nullptr, false),
       mDC(aDC),
       mContext(aContext),
       mWnd(aWindow),
       mPBuffer(nullptr),
       mPixelFormat(0) {}
 
-GLContextWGL::GLContextWGL(CreateContextFlags flags, const SurfaceCaps& caps,
-                           bool isOffscreen, HANDLE aPbuffer, HDC aDC,
+GLContextWGL::GLContextWGL(const GLContextDesc& desc, HANDLE aPbuffer, HDC aDC,
                            HGLRC aContext, int aPixelFormat)
-    : GLContext(flags, caps, nullptr, isOffscreen),
+    : GLContext(desc, nullptr, false),
       mDC(aDC),
       mContext(aContext),
       mWnd(nullptr),
@@ -320,11 +316,6 @@ void GLContextWGL::GetWSIInfo(nsCString* const out) const {
   out->Append(sWGLLib.mSymbols.fGetExtensionsStringARB(mDC));
 }
 
-already_AddRefed<GLContext> GLContextProviderWGL::CreateWrappingExisting(
-    void*, void*) {
-  return nullptr;
-}
-
 HGLRC
 WGLLibrary::CreateContextWithFallback(const HDC dc,
                                       const bool tryRobustBuffers) const {
@@ -388,7 +379,7 @@ static RefPtr<GLContext> CreateForWidget(const HWND window,
                                         LOCAL_WGL_FULL_ACCELERATION_ARB,
                                         0};
     const int* attribs;
-    if (wr::RenderThread::IsInRenderThread()) {
+    if (isWebRender) {
       attribs = kAttribsForWebRender;
     } else {
       attribs = kAttribs;
@@ -420,7 +411,7 @@ static RefPtr<GLContext> CreateForWidget(const HWND window,
                                         0};
 
     const int* attribs;
-    if (wr::RenderThread::IsInRenderThread()) {
+    if (isWebRender) {
       attribs = kAttribsForWebRender;
     } else {
       attribs = kAttribs;
@@ -441,9 +432,7 @@ static RefPtr<GLContext> CreateForWidget(const HWND window,
   const auto context = sWGLLib.CreateContextWithFallback(dc, false);
   if (!context) return nullptr;
 
-  SurfaceCaps caps = SurfaceCaps::ForRGBA();
-  const RefPtr<GLContextWGL> gl = new GLContextWGL(
-      CreateContextFlags::NONE, SurfaceCaps::ForRGBA(), false, dc, context);
+  const RefPtr<GLContextWGL> gl = new GLContextWGL({}, dc, context);
   cleanupDc.release();
   gl->mIsDoubleBuffered = true;
   if (!gl->Init()) return nullptr;
@@ -452,20 +441,20 @@ static RefPtr<GLContext> CreateForWidget(const HWND window,
 }
 
 already_AddRefed<GLContext> GLContextProviderWGL::CreateForCompositorWidget(
-    CompositorWidget* aCompositorWidget, bool aWebRender,
+    CompositorWidget* aCompositorWidget, bool aHardwareWebRender,
     bool aForceAccelerated) {
   if (!aCompositorWidget) {
     MOZ_ASSERT(false);
     return nullptr;
   }
-  return CreateForWidget(aCompositorWidget->AsWindows()->GetHwnd(), aWebRender,
-                         aForceAccelerated)
+  return CreateForWidget(aCompositorWidget->AsWindows()->GetHwnd(),
+                         aHardwareWebRender, aForceAccelerated)
       .forget();
 }
 
 /*static*/
 already_AddRefed<GLContext> GLContextProviderWGL::CreateHeadless(
-    const CreateContextFlags flags, nsACString* const out_failureId) {
+    const GLContextCreateDesc& desc, nsACString* const out_failureId) {
   auto& wgl = sWGLLib;
   if (!wgl.EnsureInitialized()) return nullptr;
 
@@ -509,10 +498,9 @@ already_AddRefed<GLContext> GLContextProviderWGL::CreateHeadless(
   const auto context = wgl.CreateContextWithFallback(dc, true);
   if (!context) return nullptr;
 
-  const bool isOffscreen = true;
+  const auto fullDesc = GLContextDesc{desc, true};
   const RefPtr<GLContextWGL> gl =
-      new GLContextWGL(flags, SurfaceCaps::Any(), isOffscreen, pbuffer, dc,
-                       context, chosenFormat);
+      new GLContextWGL(fullDesc, pbuffer, dc, context, chosenFormat);
   cleanupPbuffer.release();
   cleanupDc.release();
   if (!gl->Init()) return nullptr;
@@ -520,20 +508,6 @@ already_AddRefed<GLContext> GLContextProviderWGL::CreateHeadless(
   return RefPtr<GLContext>(gl.get()).forget();
 }
 
-/*static*/
-already_AddRefed<GLContext> GLContextProviderWGL::CreateOffscreen(
-    const IntSize& size, const SurfaceCaps& minCaps, CreateContextFlags flags,
-    nsACString* const out_failureId) {
-  *out_failureId = NS_LITERAL_CSTRING("FEATURE_FAILURE_WGL_INIT");
-
-  RefPtr<GLContext> gl = CreateHeadless(flags, out_failureId);
-  if (!gl) return nullptr;
-
-  if (!gl->InitOffscreen(size, minCaps)) return nullptr;
-
-  return gl.forget();
-}
-
 /*static*/
 GLContext* GLContextProviderWGL::GetGlobalContext() { return nullptr; }
 
