@@ -46,6 +46,8 @@ AutoInitializeImageLib::AutoInitializeImageLib() {
   // Ensure AVIF is enabled to run decoder tests.
   rv = Preferences::SetBool("image.avif.enabled", true);
   EXPECT_TRUE(rv == NS_OK);
+  rv = Preferences::SetBool("image.avif.sequence.enabled", true);
+  EXPECT_TRUE(rv == NS_OK);
 
 #ifdef MOZ_JXL
   // Ensure JXL is enabled to run decoder tests.
@@ -77,13 +79,13 @@ void ImageBenchmarkBase::SetUp() {
   // Figure out how much data we have.
   uint64_t length;
   nsresult rv = inputStream->Available(&length);
-  ASSERT_TRUE(NS_SUCCEEDED(rv));
+  ASSERT_NS_SUCCEEDED(rv);
 
   // Write the data into a SourceBuffer.
   mSourceBuffer = new SourceBuffer();
   mSourceBuffer->ExpectLength(length);
   rv = mSourceBuffer->AppendFromInputStream(inputStream, length);
-  ASSERT_TRUE(NS_SUCCEEDED(rv));
+  ASSERT_NS_SUCCEEDED(rv);
   mSourceBuffer->Complete(NS_OK);
 }
 
@@ -133,7 +135,7 @@ void SpinPendingEvents() {
   do {
     processed = false;
     nsresult rv = mainThread->ProcessNextEvent(false, &processed);
-    EXPECT_TRUE(NS_SUCCEEDED(rv));
+    EXPECT_NS_SUCCEEDED(rv);
   } while (processed);
 }
 
@@ -216,11 +218,18 @@ bool RectIsSolidColor(SourceSurface* aSurface, const IntRect& aRect,
             abs(pmColor.mGreen - gotColor.mGreen) > aFuzz ||
             abs(pmColor.mRed - gotColor.mRed) > aFuzz ||
             abs(pmColor.mAlpha - gotColor.mAlpha) > aFuzz) {
-          EXPECT_EQ(pmColor.mBlue, gotColor.mBlue);
-          EXPECT_EQ(pmColor.mGreen, gotColor.mGreen);
-          EXPECT_EQ(pmColor.mRed, gotColor.mRed);
-          EXPECT_EQ(pmColor.mAlpha, gotColor.mAlpha);
-          ASSERT_EQ_OR_RETURN(expectedPixel, gotPixel, false);
+          EXPECT_EQ(expectedPixel, gotPixel)
+              << "Color mismatch for rectangle from " << aRect.TopLeft()
+              << " to " << aRect.BottomRight() << ": "
+              << "got rgba(" << static_cast<int>(gotColor.mRed) << ", "
+              << static_cast<int>(gotColor.mGreen) << ", "
+              << static_cast<int>(gotColor.mBlue) << ", "
+              << static_cast<int>(gotColor.mAlpha) << "), "
+              << "expected rgba(" << static_cast<int>(pmColor.mRed) << ", "
+              << static_cast<int>(pmColor.mGreen) << ", "
+              << static_cast<int>(pmColor.mBlue) << ", "
+              << static_cast<int>(pmColor.mAlpha) << ")";
+          return false;
         }
       }
     }
@@ -747,6 +756,11 @@ ImageTestCase GreenFirstFrameAnimatedWebPTestCase() {
                        IntSize(100, 100), TEST_CASE_IS_ANIMATED);
 }
 
+ImageTestCase GreenFirstFrameAnimatedAVIFTestCase() {
+  return ImageTestCase("first-frame-green.avif", "image/avif",
+                       IntSize(100, 100), TEST_CASE_IS_ANIMATED);
+}
+
 ImageTestCase BlendAnimatedGIFTestCase() {
   return ImageTestCase("blend.gif", "image/gif", IntSize(100, 100),
                        TEST_CASE_IS_ANIMATED);
@@ -762,6 +776,11 @@ ImageTestCase BlendAnimatedWebPTestCase() {
                        TEST_CASE_IS_TRANSPARENT | TEST_CASE_IS_ANIMATED);
 }
 
+ImageTestCase BlendAnimatedAVIFTestCase() {
+  return ImageTestCase("blend.avif", "image/avif", IntSize(100, 100),
+                       TEST_CASE_IS_TRANSPARENT | TEST_CASE_IS_ANIMATED);
+}
+
 ImageTestCase CorruptTestCase() {
   return ImageTestCase("corrupt.jpg", "image/jpeg", IntSize(100, 100),
                        TEST_CASE_HAS_ERROR);
@@ -923,7 +942,8 @@ ImageTestCase DownscaledTransparentICOWithANDMaskTestCase() {
 }
 
 ImageTestCase TruncatedSmallGIFTestCase() {
-  return ImageTestCase("green-1x1-truncated.gif", "image/gif", IntSize(1, 1));
+  return ImageTestCase("green-1x1-truncated.gif", "image/gif", IntSize(1, 1),
+                       TEST_CASE_IS_TRANSPARENT);
 }
 
 ImageTestCase LargeICOWithBMPTestCase() {
