@@ -15,8 +15,27 @@
  */
 
 const versionsPerRelease = new Map([
-  // This is a mapping from Chromium version => Puppeteer version.
-  // In Chromium roll patches, use 'NEXT' for the Puppeteer version.
+  // This is a mapping from Chrome version => Puppeteer version.
+  // In Chrome roll patches, use `NEXT` for the Puppeteer version.
+  ['113.0.5672.63', 'v20.1.0'],
+  ['112.0.5615.121', 'v20.0.0'],
+  ['112.0.5614.0', 'v19.8.0'],
+  ['111.0.5556.0', 'v19.7.0'],
+  ['110.0.5479.0', 'v19.6.0'],
+  ['109.0.5412.0', 'v19.4.0'],
+  ['108.0.5351.0', 'v19.2.0'],
+  ['107.0.5296.0', 'v18.1.0'],
+  ['106.0.5249.0', 'v17.1.0'],
+  ['105.0.5173.0', 'v15.5.0'],
+  ['104.0.5109.0', 'v15.1.0'],
+  ['103.0.5059.0', 'v14.2.0'],
+  ['102.0.5002.0', 'v14.0.0'],
+  ['101.0.4950.0', 'v13.6.0'],
+  ['100.0.4889.0', 'v13.5.0'],
+  ['99.0.4844.16', 'v13.2.0'],
+  ['98.0.4758.0', 'v13.1.0'],
+  ['97.0.4692.0', 'v12.0.0'],
+  ['93.0.4577.0', 'v10.2.0'],
   ['92.0.4512.0', 'v10.0.0'],
   ['91.0.4469.0', 'v9.0.0'],
   ['90.0.4427.0', 'v8.0.0'],
@@ -39,4 +58,16 @@ const versionsPerRelease = new Map([
   ['73.0.3679.0', 'v1.12.2'],
 ]);
 
-module.exports = versionsPerRelease;
+// Should not be more than 2 major versions behind Chrome Stable (https://chromestatus.com/roadmap).
+const lastMaintainedChromeVersion = '109.0.5412.0';
+
+if (!versionsPerRelease.has(lastMaintainedChromeVersion)) {
+  throw new Error(
+    'lastMaintainedChromeVersion is missing from versionsPerRelease'
+  );
+}
+
+module.exports = {
+  versionsPerRelease,
+  lastMaintainedChromeVersion,
+};
